"""
The iterated prisoner's dilemma and the State Pattern.

written by Niko Wilbert for the G-Node Python Winter School 2010

from Wikipedia:

Two suspects are arrested by the police. The police have insufficient evidence
for a conviction, and, having separated both prisoners, visit each of them to
offer the same deal. If one testifies (defects from the other) for the
prosecution against the other and the other remains silent (cooperates with the
other), the betrayer goes free and the silent accomplice receives the
full 10-year sentence. If both remain silent, both prisoners are sentenced to
only 3 years in jail for a minor charge. If each betrays the other, each
receives a five-year sentence. Each prisoner must choose to betray the other
or to remain silent. Each one is assured that the other would not know about
the betrayal before the end of the investigation. How should the prisoners act?

In the iterated prisoner's dilemma, the game is played repeatedly. Thus each
player has an opportunity to punish the other player for previous
non-cooperative play.
"""

import random

### The Rules ###

# table with the years the prisoners get in each case
penalties = {('silent', 'silent'): (3, 3),
             ('testify', 'silent'): (0, 10),
             ('silent', 'testify'): (10, 0),
             ('testify', 'testify'): (5, 5)}

def play_game(prisoner1, prisoner2, n_rounds=100):
    last_reactions = ('first round', 'first round')  # tuple can be hashed
    reactions = [None, None]
    prisoners = [prisoner1, prisoner2]
    for round in range(n_rounds):
        reactions[0] = prisoners[0].testify(last_reactions[1])
        reactions[1] = prisoners[1].testify(last_reactions[0])
        last_reactions = tuple(reactions)
        prisoners[0].years += penalties[last_reactions][0]
        prisoners[1].years += penalties[last_reactions][1]
        print ("round %03d: %s --> %s, %s" %
               (round, str(last_reactions),
                prisoners[0].get_years(), prisoners[1].get_years()))
    if prisoners[0].years < prisoners[1].years:
        print "The first player wins."  
    elif prisoners[0].years > prisoners[1].years:
        print "The second player wins."
    else:
        print "We have a draw!"  
        

### State Pattern ###

class PrisonerState(object):
    """Base class for prisoner states."""
    
    def __init__(self, agent=None):
        """Initialize this state.
        
        agent -- The agent instance to which this state belongs.
        """
        self.agent = agent
    
    def testify(self, previous_reaction):
        """Return reaction of the prisoner.
        
        previous_reaction -- What the other prisoner did last time.
        """
        raise NotImplementedError()


class CooperativeState(PrisonerState):
    """State representing defensive behavior.
    
    Cooperate unless the other prisoner testifies against us repeatedly.
    """
    
    def __init__(self, agent=None, tolerance=2):
        """Initialize this state.
        
        agent -- The agent instance to which this state belongs.
        tolerance -- Number to specify after how many testifications of the
            other prisoner the state is changed to the aggresive state.
        """
        super(CooperativeState, self).__init__(agent)
        self._tolerance = tolerance
        self._n_testifications = 0
    
    def testify(self, previous_reaction):
        if previous_reaction == 'first round':
            return 'silent'
        if previous_reaction == 'testify':
            self._n_testifications += 1
        if self._n_testifications > self._tolerance:
            self._n_testifications = 0
            self.agent.state = self.agent.states['aggressive']
            return 'testify'
        else:
            return 'silent'
            
            
class AggressiveState(PrisonerState):
    """State representing defensive behavior.
    
    We testify unless the other prisoner cooperates, after which we cooperate
    as well and switch to the cooperative state.
    """
    
    def testify(self, previous_reaction):
        if previous_reaction == 'first round':
            return 'testify'
        if previous_reaction == 'silent':
            self.agent.state = self.agent.states['cooperative']
        return 'testify' 


class PrisonerAgent(object):
    """Class representing a prisoner.
    
    This class is the context for the states.
    """
    
    def __init__(self, start_key='cooperative', states=None):
        """Initialize the internal state.
        
        start_key -- Key of the initial state (default is 'cooperative').
        states -- Dict with PrisonerState instances. 
        """
        self.years = 0
        if states:
            for state in states.values():
                state.agent = self
            self.states = states
        else:
            self.states = {'cooperative': CooperativeState(self),
                           'aggressive': AggressiveState(self)}
        self.state = self.states[start_key]
        
    def testify(self, previous_reaction):
        """Return reaction of the prisoner.
        
        previous_reaction -- What the other prisoner did last time, values
            are 'silent', 'testify', and 'first round'. 
        """
        return self.state.testify(previous_reaction)
    
    def get_years(self):
        """Return a string with the number of years."""
        return str(self.years) + " years"
    

## Example usage ##

# a modified state to demonstrate the flexibility

class UnstableState(PrisonerState):
    """State representing unstable cooperative behaviour.
    
    The state will randomly switch to 'aggresive'.
    """
    
    def testify(self, previous_reaction):
        if random.random() >= 0.5:
            self.agent.state = self.agent.states['aggressive']
        return 'silent'


if __name__ == '__main__':
    prisoner1 = PrisonerAgent()
    prisoner2 = PrisonerAgent(states={'cooperative': UnstableState(),
                                      'aggressive': AggressiveState()})
    play_game(prisoner1, prisoner2, 20)
 
        